#
# Copyright (c) 2003, Ashok P. Nadkarni
# All rights reserved.
#
# See the file LICENSE for license

# TBD - convert file spec to drive root path
package require cmdline
package require twapi
#package require trycatch 2.0
#namespace import trycatch::*
package require TclExcept

#
# Get info associated with a drive
proc twapi::get_drive_info {drive args} {
    variable windefs

    set drive [_drive_rootpath $drive]
    
    array set opts [cmdline::getoptions args {
        all size freespace used useravail type serialnum volumename maxcomponentlen fstype attr
    }]
    if {$opts(all)} {
        array_set_all opts 1
    }

    set result [list ]
    if {$opts(size) || $opts(freespace) || $opts(used) || $opts(useravail)} {
        foreach {useravail size freespace} [GetDiskFreeSpaceEx $drive] {break}
        foreach opt {size freespace useravail}  {
            if {$opts($opt)} {
                lappend result -$opt [set $opt]
            }
        }
        if {$opts(used)} {
            lappend result -used [expr {$size - $freespace}]
        }
    }

    if {$opts(type)} {
        lappend result -type [get_drive_type $drive]
    }

    if {$opts(serialnum) || $opts(volumename) || $opts(maxcomponentlen)
        || $opts(fstype) || $opts(attr)} {
        foreach {volumename serialnum maxcomponentlen attr fstype} \
            [GetVolumeInformation $drive] { break }
        foreach opt {volumename maxcomponentlen fstype}  {
            if {$opts($opt)} {
                lappend result -$opt [set $opt]
            }
        }
        if {$opts(serialnum)} {
            set low [expr {$serialnum & 0x0000ffff}]
            set high [expr {($serialnum >> 16) & 0x0000ffff}]
            lappend result -serialnum [format "%.4X-%.4X" $high $low]
        }
        if {$opts(attr)} {
            set attrs [list ]
            foreach val {
                case_preserved_names
                unicode_on_disk
                persistent_acls
                file_compression
                volume_quotas
                supports_sparse_files
                supports_reparse_points
                supports_remote_storage
                volume_is_compressed
                supports_object_ids
                supports_encryption
                named_streams
                read_only_volume
            } {
                # Coincidentally, the attribute values happen to match
                # the corresponding constant defines
                set cdef "FILE_[string toupper $val]"
                if {$attr & $windefs($cdef)} {
                    lappend attrs $val
                }
            }
            lappend result -attr $attrs
        }
    }
    return $result
}


# Check if disk has at least n bytes available for the user (NOT total free)
proc twapi::user_drive_space_available {drv space} {
    return [expr {$space <= [lindex [get_drive_info $drv -useravail] 1]}]
}

# Get the drive type
proc twapi::get_drive_type {drive} {
    set type [GetDriveType "[string trimright $drive :/\\]:\\"]
    switch -exact -- $type {
        0 { return unknown}
        1 { error "Invalid drive"}
        2 { return removable}
        3 { return fixed}
        4 { return remote}
        5 { return cdrom}
        6 { return ramdisk}
    }
}

#
# Get list of drives
proc twapi::get_logical_drives {args} {
    eval set [cmdline::getoptions args {{type.arg ""}}]

    set drives [list ]

    set i 0
    set drivebits [GetLogicalDrives]
    foreach drive {A B C D E F G H I J K L M N O P Q R S T U V W X Y Z} {
        if {[expr {$drivebits & (1 << $i)}]} {
            if {([string length $type] == 0) ||
                [string equal [get_drive_type $drive] $type]} {
                lappend drives $drive:
            }
        }
        incr i
    }
    return $drives
}

#
# Set the drive label
proc twapi::set_drive_volumename {drive label} {
    SetVolumeLabel [_drive_rootpath $drive] $label
}

#
# Utility functions

proc twapi::_drive_rootpath {drive} {
    return "[string trimright $drive :/\\]:\\"
}
